"""LLM API handler that generates various text for the README.md file."""

import asyncio
import time
import traceback
from contextlib import asynccontextmanager
from typing import Any, Dict, List, Tuple

import openai
from cachetools import TTLCache
from httpx import (
    AsyncClient,
    HTTPStatusError,
    Limits,
    NetworkError,
    TimeoutException,
)
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

import readmeai.config.settings as settings
import readmeai.core.logger as logger
from readmeai.core.tokens import (
    adjust_max_tokens,
    get_token_count,
    truncate_tokens,
)
from readmeai.utils.utils import format_sentence


class LlmApiHandler:
    """LLM API handler that generates text for the README.md file."""

    @asynccontextmanager
    async def use_api(self) -> None:
        """Context manager for manage resources used by the HTTP client."""
        try:
            yield
        finally:
            await self.close()

    async def close(self) -> None:
        """Closes the HTTP client."""
        await self.http_client.aclose()

    def __init__(self, config: settings.AppConfig) -> None:
        """Initializes the GPT language model API handler."""
        self.config = config
        self.logger = logger.Logger(__name__)
        self.prompts = config.prompts
        self.tokens = config.llm.tokens
        self.tokens_max = config.llm.tokens_max
        self.cache = TTLCache(maxsize=1000, ttl=1200)
        self.http_client = AsyncClient(
            http2=True,
            timeout=20,
            limits=Limits(max_keepalive_connections=20, max_connections=200),
        )
        self.last_request_time = time.monotonic()
        self.rate_limit_semaphore = asyncio.Semaphore(config.llm.rate_limit)

    async def prompt_processor(
        self, prompt_type: str, context: Dict[str, Any]
    ) -> Dict[str, str]:
        """Processes a prompt and returns the generated text.

        Parameters
        ----------
        prompt_type
            The type of prompt to generate i.e. features, overview, slogan.
        context
            A dictionary containing the context for the prompt.

        Returns
        -------
        Dict[str, str]
            A dictionary containing the prompt type and the generated text.
        """
        self.logger.info(f"Processing prompt: {prompt_type}")

        if prompt_type == "summaries":
            return await self._generate_summaries(context)

        formatted_prompt = self.generate_prompt(prompt_type, context)
        tokens = adjust_max_tokens(self.tokens, formatted_prompt)
        _, summary = await self.generate_text(
            prompt_type, formatted_prompt, tokens
        )
        return {prompt_type: summary}

    def generate_prompt(self, prompt_type, context) -> str:
        """Generates a prompt for the LLM API.

        Parameters
        ----------
        prompt_type
            The type of prompt to generate i.e. features, overview, slogan.
        context
            A dictionary containing the context for the prompt.

        Returns
        -------
        str
            The generated prompt.
        """
        prompt_templates = {
            "features": self.prompts.features,
            "overview": self.prompts.overview,
            "slogan": self.prompts.slogan,
        }
        return prompt_templates[prompt_type].format(
            *[context[key] for key in sorted(context.keys())]
        )

    async def _generate_summaries(
        self, file_context: List[Tuple[str, str]]
    ) -> List[Tuple[str, str]]:
        """Generates code summaries for the README.md file.

        Parameters
        ----------
        file_context
            A list of tuples containing the file path and file content.

        Returns
        -------
        List[Tuple[str, str]]
            List of tuples containing the file path and the generated summary.
        """
        code_summaries = []

        for context in file_context["summaries"]:
            file_path, file_content = context
            prompt = self.prompts.summaries.format(
                self.config.md.tree, file_path, file_content
            )
            tokens = adjust_max_tokens(self.tokens, prompt)
            _, summary = await self.generate_text(file_path, prompt, tokens)
            code_summaries.append((file_path, summary))

        return code_summaries

    @retry(
        stop=stop_after_attempt(4),
        wait=wait_exponential(multiplier=1, min=2, max=6),
        retry=retry_if_exception_type(
            (
                openai.error.OpenAIError,
                HTTPStatusError,
                NetworkError,
                TimeoutException,
            )
        ),
    )
    async def generate_text(
        self, index: str, prompt: str, tokens: int
    ) -> Tuple[str, str]:
        """Generates text for the README.md file using GPT language models.

        Parameters
        ----------
        index
            Type of prompt i.e. features, overview, slogan, or code summary.
        prompt
            The prompt to generate text from.
        tokens
            The number of tokens to generate.

        Returns
        -------
        Tuple[str, str]
            A tuple containing the prompt type and the generated text.
        """
        token_count = get_token_count(prompt, self.config.llm.encoding)
        if token_count > self.tokens_max:
            self.logger.warning(
                f"Truncating tokens: {token_count} > {self.tokens_max} max"
            )
            prompt = truncate_tokens(prompt, tokens)

        try:
            async with self.rate_limit_semaphore:
                response = await self.http_client.post(
                    self.config.llm.endpoint,
                    headers={"Authorization": f"Bearer {openai.api_key}"},
                    json={
                        "messages": [
                            {
                                "role": "system",
                                "content": self.config.llm.content,
                            },
                            {"role": "user", "content": prompt},
                        ],
                        "model": self.config.llm.model,
                        "temperature": self.config.llm.temperature,
                        "max_tokens": tokens,
                    },
                )
                response.raise_for_status()
                llm_text = response.json()
                summary = llm_text["choices"][0]["message"]["content"]
                summary = (
                    format_sentence(summary)
                    if index != "features"
                    else summary
                )
                self.cache[prompt] = summary
                self.logger.info(f"Response: {index} - {summary}")

                return index, summary

        except Exception as exc_info:
            self.logger.error(f"Exception making request: {exc_info}")
            return index, f"{exc_info}: {traceback.format_exc()}"



"""Utility helper functions for the README-AI package."""

import re
from importlib import resources
from pathlib import Path
from typing import List

from pkg_resources import resource_filename

from readmeai.config.settings import ConfigHelper
from readmeai.core import logger

logger = logger.Logger(__name__)


def is_valid_url(url: str) -> bool:
    """Validate a URL string.

    Parameters
    ----------
    url
        The URL string to validate.

    Returns
    -------
        True if the URL is valid, False otherwise.
    """
    regex = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain
        r"localhost|"  # localhost
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|"  # ipv4
        r"\[?[A-F0-9]*:[A-F0-9:]+\]?)"  # ipv6
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )
    return re.match(regex, url) is not None


def flatten_list(nested_list: List[List]) -> List:
    """Flatten a nested list (list of lists converted to a single list).

    Parameters
    ----------
    nested_list
        The nested list to flatten.

    Returns
    -------
        A flattened list.
    """
    result = []
    for item in nested_list:
        if isinstance(item, list):
            result.extend(flatten_list(item))
        else:
            result.append(item)
    return result


def format_sentence(text: str) -> str:
    """Formats output text from the LLM model.
    i.e. removes extra white space, newlines, tabs, etc.

    Parameters
    ----------
    text
        The text to format.

    Returns
    -------
        The formatted text, used to generate the README.
    """
    # Remove single and double quotes around any text
    text = re.sub(r"(?<!\w)['\"](.*?)['\"](?!\w)", r"\1", text)

    # Remove newlines and tabs
    text = text.replace("\n", "").replace("\t", "")

    # Remove non-letter characters from the beginning of the string
    text = re.sub(r"^[^a-zA-Z]*", "", text)

    # Remove extra white space around punctuation except for '('
    text = re.sub(r"\s*([)'.!,?;:])(?!\.\s*\w)", r"\1", text)

    # Remove extra white space before opening parentheses
    text = re.sub(r"(\()\s*", r"\1", text)

    # Replace multiple consecutive spaces with a single space
    text = re.sub(r" +", " ", text)

    # Remove extra white space around hyphens
    text = re.sub(r"\s*-\s*", "-", text)

    return text.strip()


def get_relative_path(absolute_path: str, base_path: str) -> str:
    """Get the relative path of a file.

    Parameters
    ----------
    absolute_path
        Absolute path to the file i.e. the full path on the local machine.
    base_path
        Base path to use for the relative path i.e. the project root.

    Returns
    -------
        The relative path string of the file.
    """
    absolute_path = Path(absolute_path)
    return absolute_path.relative_to(base_path)


def get_resource_path(package: str, resource_name: str) -> Path:
    """Use importlib.resources or pkg_resources to get resource path.
    Python 3.9+ uses importlib.resources, older versions use pkg_resources.

    Parameters
    ----------
    package
        Python package name.
    resource_name
        Name of the resource to get the path for.

    Returns
    -------
        The path to the resource file.
    """
    try:
        resource_path = resources.files(package) / resource_name
    except (TypeError, FileNotFoundError):
        resource_path = resource_filename(package, resource_name)
    return resource_path


def remove_substring(
    input_string: str, pattern: str = r"</p>.*?</div>"
) -> str:
    """Remove a substring from a string.

    Parameters
    ----------
    input_string
        The string to remove the substring from.
    pattern, optional
        The substring to remove, by default r"</p>.*?</div>".

    Returns
    -------
        The input string with the substring removed.
    """
    output_string = re.sub(
        pattern, "</p>\n</div>", input_string, flags=re.DOTALL
    )
    return output_string


def should_ignore(conf_helper: ConfigHelper, file_path: Path) -> bool:
    """Check if a file should be ignored and not passed to the LLM API.
    Uses a default list of files - 'readmeai/settings/ignore_files.toml'.

    Parameters
    ----------
    conf_helper
        Configuration helper instance with the ignore files list.
    file_path
        The path to the file to check.

    Returns
    -------
        True if the file should be ignored, False otherwise.
    """
    ignore_files = conf_helper.ignore_files

    if (
        (file_path.name in ignore_files["files"])
        or (file_path.suffix.lstrip(".") in ignore_files["extensions"])
        or any(
            directory in file_path.parts
            for directory in ignore_files["directories"]
        )
    ):
        logger.debug(f"Ignoring item: {file_path.name}")
        return True

    return False



"""Git service utilities to retrieve repository metadata."""

import os
import re
from pathlib import Path
from typing import Tuple

from readmeai.config.settings import GitService


def get_remote_file_url(file_path: str, full_name: str, repo_url: str) -> str:
    """Returns the URL of the file in the remote repository."""
    if Path(repo_url).exists():
        return GitService.LOCAL.file_url.format(file_path=file_path)

    for service in GitService:
        if service.host in repo_url:
            return service.file_url.format(
                full_name=full_name, file_path=file_path
            )

    raise ValueError("Unsupported Git service for URL generation.")


def get_remote_full_name(url_or_path) -> Tuple[str, str]:
    """Returns the full name of the repository."""
    if os.path.exists(url_or_path):
        return "local", os.path.basename(url_or_path)

    patterns = {
        GitService.GITHUB.host: r"https?://github.com/([^/]+)/([^/]+)",
        GitService.GITLAB.host: r"https?://gitlab.com/([^/]+)/([^/]+)",
        GitService.BITBUCKET.host: r"https?://bitbucket.org/([^/]+)/([^/]+)",
    }

    for host, pattern in patterns.items():
        match = re.match(pattern, url_or_path)
        if match:
            username, repo_name = match.groups()
            return username, repo_name

    raise ValueError("Error: invalid repository URL or path.")



"""Parser for Rust dependency files."""

from typing import List

import toml

from readmeai.parsers.base_parser import FileParser

TOM_DECODE_ERROR = "Error decoding TOML content: {0}"


class CargoTomlParser(FileParser):
    """
    Parser for Rust cargo.toml dependency files.
    """

    def parse(self, content: str) -> List[str]:
        """Extract packages names from Rust TOML files."""
        try:
            data = toml.loads(content)
            dependencies = []
            if "dependencies" in data:
                dependencies.extend(data.get("dependencies", {}).keys())
            if "dev-dependencies" in data:
                dependencies.extend(data.get("dev-dependencies", {}).keys())
            return dependencies

        except toml.TomlDecodeError as error:
            self.log_error(TOM_DECODE_ERROR.format(error))
            return dependencies



"""Creates a directory tree structure for a code repository."""

from pathlib import Path

from readmeai.config.settings import ConfigHelper
from readmeai.core import logger
from readmeai.utils import utils

logger = logger.Logger(__name__)


class TreeGenerator:
    """Generates a directory tree structure for a code repository."""

    def __init__(
        self,
        conf_helper: ConfigHelper,
        root_directory: Path,
        repo_url: str,
        repo_name: str,
        max_depth: int = 3,
    ):
        self.config_helper = conf_helper
        self.root_directory = root_directory
        self.repo_name = repo_name
        self.repo_url = repo_url
        self.max_depth = max_depth

    def generate_and_format_tree(self) -> str:
        """Generates and formats a tree structure."""
        tree_str = self._generate_tree(self.root_directory)
        formatted_tree_str = self._format_tree(tree_str)
        return formatted_tree_str

    def _generate_tree(
        self,
        directory: Path,
        prefix: str = "",
        is_last: bool = True,
        parent_prefix: str = "",
        depth: int = 0,
    ) -> str:
        """Generates a tree structure for a given directory."""
        if depth > self.max_depth:
            return ""

        if utils.should_ignore(self.config_helper, directory):
            return ""

        display_name = "." if directory == self.repo_url else directory.name
        box_branch = "└── " if is_last else "├── "
        parts = [parent_prefix + box_branch + display_name]

        if directory.is_dir():
            parts.append("/\n")
            children = sorted(
                [
                    child
                    for child in directory.iterdir()
                    if child.name != ".git"
                ]
            )
            for index, child in enumerate(children):
                is_last_child = index == len(children) - 1
                child_prefix = "    " if is_last else "│   "
                parts.append(
                    self._generate_tree(
                        child,
                        box_branch,
                        is_last_child,
                        f"{parent_prefix}{child_prefix}",
                        depth + 1,
                    )
                )
        else:
            parts.append("\n")

        return "".join(parts)

    def _format_tree(self, tree_str: str) -> str:
        """Formats the directory tree structure."""
        tree_str = tree_str.split("\n", 1)
        tree_str[0] = f"└── {self.repo_name}/"
        return "\n".join(tree_str)
